RKDRE_se <- function(data, formula, tol_beta = 1e-12, tol_EM = 1e-12, max.iter = 100, 
                     bandwidth = c("normal","robust1","robust2","SJ","bcv","ucv"), update_bw = FALSE, std_error = FALSE, r = 0.5, a = 3, 
                     h.SJ = FALSE){
  
  n <- nrow(data)
  resid <- NULL
  h <- NULL
  std_error_trimmed <- NULL
  std_error_untrimmed <- NULL
  
  #compute OLS estimate
  OLS_model <- lm(data = data, formula = formula)
  beta_OLS <- OLS_model$coefficients
  
  #set initial estimate
  beta <- beta_OLS
  
  #set initial bandwidth
  if (bandwidth == "normal"){
    h <- 1.06 * sd(OLS_model$residuals) * n^(-0.2)
  } else if (bandwidth == "robust1"){
    h <- 0.79 * IQR(OLS_model$residuals) * n^(-0.2)
  } else if (bandwidth == "robust2"){
    h <- 0.9 * min(sd(OLS_model$residuals),IQR(OLS_model$residuals)/1.34) * n^(-0.2)
  } else if (bandwidth == "SJ"){
    h <- bw.SJ(OLS_model$residuals)
  } else if (bandwidth == "bcv"){
    h <- bw.bcv(OLS_model$residuals)
  } else if (bandwidth == "ucv"){
    h <- bw.ucv(OLS_model$residuals)
  }
  
  #create matrix of explanatory variables with intercept
  X <- model.matrix(formula, as.data.frame(data))
  
  #column of dependent variable
  y <- data[,which(colnames(data) == all.vars(formula)[1])]
  
  #create A for in EM loop
  A <- solve( t(X) %*% X )
  
  converged_beta <- 0 
  counter <- 0
  
  while (converged_beta == 0 && counter < max.iter){ #start beta loop
    
    counter <- counter + 1
    beta_prev <- beta
    beta_EM <- beta_prev
    converged_EM <- 0
    
    #create residuals
    resid <- y - X %*% beta_prev
    
    if (update_bw == TRUE){
      #update bandwidth
      if (bandwidth == "normal"){
        h <- 1.06 * sd(resid) * n^(-0.2)
      } else if (bandwidth == "robust1"){
        h <- 0.79 * IQR(resid) * n^(-0.2)
      } else if (bandwidth == "robust2"){
        h <- 0.9 * min(sd(resid),IQR(resid)/1.34) * n^(-0.2)
      } else if (bandwidth == "SJ"){
        h <- bw.SJ(resid)
      } else if (bandwidth == "bcv"){
        h <- bw.bcv(resid)
      } else if (bandwidth == "ucv"){
        h <- bw.ucv(resid)
      }
    }
    
    resid_matrix <- do.call(rbind, replicate(n, t(resid), simplify=FALSE))
    
    while (converged_EM == 0){ #start EM loop
      
      beta_prev_EM <- beta_EM
      
      resid_EM <- y - X %*% beta_prev_EM
      
      p_matrix <- matrix(rep(resid_EM,n), nrow = n, ncol = n) - resid_matrix
      
      p_matrix <- exp( -0.5 * (p_matrix / h )^2 )
      
      p_matrix <- p_matrix / ( rowSums(p_matrix) )
      
      #split beta_update in two parts (beta_OLS and A %*% (B-C))
      #A is solve( t(X) %*% X) is created outside the loops
      
      PE_sum <- p_matrix %*% resid 
      
      B <- colSums(X * as.vector(PE_sum))
      
      C <- mean(PE_sum) * colSums(X)
      
      #update beta
      beta_EM <- beta_OLS - A %*% (B - C)
      
      if(max((beta_EM - beta_prev_EM)^2) < tol_EM){
        converged_EM <- 1
        beta <- beta_EM
      }
      
    } #end of EM loop
    
    if (max((beta - beta_prev)^2) < tol_beta){
      converged_beta <- 1
      resid <- y - X %*% beta
    }
  } #end of beta loop
  
  if (counter >= max.iter){
    warning('RKDRE did not converge in specified maximum number of iterations')
  }
  
  if (h.SJ == TRUE){
    h <- bw.SJ(resid, method = "ste", tol = 0.00001 * 1.144 * min(sd(resid),IQR(resid)/1.34) * n^(-0.2))
  }
  
  score_adaptive <- akj(x = as.vector(resid), z = as.vector(resid), h = h, iker1 = 1)
  
  ############################################################################################
  #compute standard error via traditional kernel estimate using first derivative only: SE1(T)
  ############################################################################################
  #create matrix for score computation
  resid_matrix <- ( matrix(rep(resid,n), nrow = n, ncol = n) - do.call(rbind, replicate(n, t(resid), simplify=FALSE)) ) / h
  
  exp_resid_matrix <- exp( -0.5 * ( resid_matrix )^2 )
  
  #compute kernel density (up to scale factor)
  f <- 1/(n * h * sqrt(2*pi)) * rowSums( exp_resid_matrix )
  
  #compute first derivative of kernel density (up to scale factor)
  df <- -1/(n * h^2 * sqrt(2*pi)) * rowSums( resid_matrix * exp_resid_matrix ) 
  
  #compute score function
  psi_kernel <-  df / f 
  untrimmed1 <- X * psi_kernel
  SE1 <- sqrt( diag( solve( t(untrimmed1) %*% untrimmed1) ) )
  
 
  ############################################################################################
  #compute standard error via traditional kernel estimate using first and second derivatives: SE2
  ############################################################################################
  #compute quotient second derivative of kernel density and kernel density (fn''(x)/fn(x))
  ddf_f <- 1/(n*h^3*sqrt(2*pi)) * rowSums ( resid_matrix^2 * exp_resid_matrix) / f - rep(1/(h^2),n)
  score_kernel <- psi_kernel^2 - ddf_f
  p <- ncol(X)
  #create 3d array of outer products of the rows of X
  X_outer <- array(apply(X,1,function(x){x %*% t(x)}),dim=c(p,p,n))
  untrimmed2 <- apply(sweep(X_outer,3,FUN = "*",score_kernel),c(1,2),sum)
  SE2 <- sqrt(diag(solve(untrimmed2)))
  
  
  return(list(coefficients = beta, SE1 = SE1, SE2 = SE2))
}

#function to perform a simulation
sim_std_error <- function(nrep, n = 100, nvar = 1, coef = c(1,-1),  
                          distribution = c("normal","uniform","heavy_tail","contam_normal","multi_modal", "skewed", 
                                           "heterosked_normal","unif_normal", "log_normal","gamma","truncnorm","beta"),
                          tol_beta = 1e-9, tol_EM = 1e-9, bandwidth = c("normal","robust1","robust2","SJ","bcv","ucv"),
                          update_bw = FALSE, h.SJ = TRUE){
  
  set.seed(123)
  #######################################
  #run pbreplicate in parallel for speed
  #######################################
  
  n_cores <- detectCores()  #specify number of nodes to use
  cl <- makeCluster(n_cores) #make cluster
  
  #functions and variables needed in cluster
  var_list <- list("get.data","RKDRE","RKDRE_se","boot.RKDRE.se","nrep", "n","nvar",
                   "coef", "distribution","tol_beta","tol_EM","bandwidth", "update_bw", "h.SJ") 
  
  clusterEvalQ(cl = cl, { #libraries needed in cluster
    library(plyr)
    library(pbapply)
    library(MASS)
    library(LaplacesDemon)
    library(stats)
    library(quantreg)
    library(boot) 
  })
  clusterExport(cl = cl, var_list, envir=environment()) #specify environment as local environment of function
  
  results <- pbreplicate(nrep, {
    d <- get.data(n = n, nvar = nvar, coef = coef, distribution = distribution)           
    frmla <- as.formula('y~.')
    bandwidth_boot <<- bandwidth
    
    #estimate standard errors via information matrix
    fit.RKDRE <- RKDRE_se(data = d, formula = frmla, tol_beta = tol_beta,
                          tol_EM = tol_EM, bandwidth = bandwidth, update_bw = update_bw, std_error = std_error, h.SJ = h.SJ)
    if(n < 101){
      #estimate standard errors via bootstrap
      boot.RKDRE <- apply(boot(data = d, boot.RKDRE.se, R = 100)$t,2,sd)
    }
    else{
      boot.RKDRE <- rep(0,(nvar+1))
    }
    c(coef = fit.RKDRE$coefficients, SE1 = fit.RKDRE$SE1, SE1T = fit.RKDRE$SE1T, SE1A = fit.RKDRE$SE1A, SE1AT = fit.RKDRE$SE1AT,
      SE2 = fit.RKDRE$SE2, SE2T = fit.RKDRE$SE2T, SE2A = fit.RKDRE$SE2A, SE2AT = fit.RKDRE$SE2AT, SEB = boot.RKDRE)
  }, cl = cl)
  
  stopCluster(cl) 
  
  #######################################
  #stop cluster
  #######################################
  p <- nvar +1
  results <- t(results)
  
  #compute mean and variance for estimate 
  res <- rbind('mean' = apply(results[,1:p],2,function(x){mean(na.omit(x))}), 'var' = apply(results[,1:p],2,function(x){var(na.omit(x))}) )
  res <- rbind(res, 'RMSE' = ( sqrt((res[1,] - coef )^2 + res[2,]) ) )
  
  #compute root of mean square of all estimated standard errors
  RMSE <- sqrt(apply(results[,-c(1:p)]^2,2,function(x){mean(na.omit(x))}))
  ratio <- RMSE / res[3,]
  
  return(list(SE1 = list(RMSE1 = RMSE[1:p], ratio1 = ratio[1:p]),
              SE2 = list(RMSE2 = RMSE[(p+1):(2*p)], ratio2 = ratio[(p+1):(2*p)]),
              SEB = list(RMSEB = RMSE[(2*p+1):(3*p)], ratioB = ratio[(2*p+1):(3*p)]),
              estimates = results,summary = res))
}

boot.RKDRE.se <- function(data, index, bandwidth = bandwidth_boot){
  data_boot <- data[index,]
  coef <- RKDRE(data = data_boot, formula = as.formula('y~.'), bandwidth = bandwidth, tol_beta = 1e-12, tol_EM = 1e-12)$coefficients
  return(coef)
}

#############################################################################################
#Table SE.A
normal.se.res <- list()
nrep <- 500
nvar <- 1
coef <- c(1,-1)
for(n in c(50,100,500,1000)){
  name <- paste("n= ",n, sep="")
  print(name)
  normal.se.res[[name]] <- sim_std_error(nrep = nrep, n = n, nvar = nvar, coef = coef, distribution = "normal", bandwidth = "normal", h.SJ = TRUE)
  print(normal.se.res[[name]]$summary)
  print(normal.se.res[[name]]$SE1$ratio1)
  print(normal.se.res[[name]]$SE1A$ratio1A)
  print(normal.se.res[[name]]$SE2$ratio2)
  print(normal.se.res[[name]]$SE2A$ratio2A)
  if (n <= 100){
    print(normal.se.res[[name]]$SEB$ratioB)
  }
}
save(normal.se.res, file = "normal.se.res.RData")

#Table SE.B
contam_normal.se.res <- list()
nrep <- 500
nvar <- 1
coef <- c(1,-1)
for(n in c(50,100,500,1000)){
  name <- paste("n= ",n, sep="")
  print(name)
  contam_normal.se.res[[name]] <- sim_std_error(nrep = nrep, n = n, nvar = nvar, coef = coef, distribution = "contam_normal", bandwidth = "normal", h.SJ = TRUE)
  print(contam_normal.se.res[[name]]$summary)
  print(contam_normal.se.res[[name]]$SE1$ratio1)
  print(contam_normal.se.res[[name]]$SE1A$ratio1A)
  print(contam_normal.se.res[[name]]$SE2$ratio2)
  print(contam_normal.se.res[[name]]$SE2A$ratio2A)
  if (n <= 100){
    print(contam_normal.se.res[[name]]$SEB$ratioB)
  }
}
save(contam_normal.se.res, file = "contam_normal.se.res.RData")

#Table SE.C
heavy_tail.se.res <- list()
nrep <- 500
nvar <- 1
coef <- c(1,-1)
for(n in c(50,100,500,1000)){
  name <- paste("n= ",n, sep="")
  print(name)
  heavy_tail.se.res[[name]] <- sim_std_error(nrep = nrep, n = n, nvar = nvar, coef = coef, distribution = "heavy_tail", bandwidth = "normal", h.SJ = TRUE)
  print(heavy_tail.se.res[[name]]$summary)
  print(heavy_tail.se.res[[name]]$SE1$ratio1)
  print(heavy_tail.se.res[[name]]$SE1A$ratio1A)
  print(heavy_tail.se.res[[name]]$SE2$ratio2)
  print(heavy_tail.se.res[[name]]$SE2A$ratio2A)
  if (n <= 100){
    print(heavy_tail.se.res[[name]]$SEB$ratioB)
  }
}
save(heavy_tail.se.res, file = "heavy_tail.se.res.t2.237.RData")

#Table SE.D
multi_modal.se.res <- list()
nrep <- 500
nvar <- 1
coef <- c(1,-1)
for(n in c(50,100,500,1000)){
  name <- paste("n= ",n, sep="")
  print(name)
  multi_modal.se.res[[name]] <- sim_std_error(nrep = nrep, n = n, nvar = nvar, coef = coef, distribution = "multi_modal", bandwidth = "normal", h.SJ = TRUE)
  print(multi_modal.se.res[[name]]$summary)
  print(multi_modal.se.res[[name]]$SE1$ratio1)
  print(multi_modal.se.res[[name]]$SE1A$ratio1A)
  print(multi_modal.se.res[[name]]$SE2$ratio2)
  print(multi_modal.se.res[[name]]$SE2A$ratio2A)
  if (n <= 100){
    print(multi_modal.se.res[[name]]$SEB$ratioB)
  }
}
save(multi_modal.se.res, file = "multi_modal.se.res.RData")

#Table SE.E
uniform.se.res <- list()
nrep <- 500
nvar <- 1
coef <- c(1,-1)
for(n in c(50,100,500,1000)){
  name <- paste("n= ",n, sep="")
  print(name)
  uniform.se.res[[name]] <- sim_std_error(nrep = nrep, n = n, nvar = nvar, coef = coef, distribution = "uniform", bandwidth = "normal", h.SJ = TRUE)
  print(uniform.se.res[[name]]$summary)
  print(uniform.se.res[[name]]$SE1$ratio1)
  print(uniform.se.res[[name]]$SE1A$ratio1A)
  print(uniform.se.res[[name]]$SE2$ratio2)
  print(uniform.se.res[[name]]$SE2A$ratio2A)
  if (n <= 100){
    print(uniform.se.res[[name]]$SEB$ratioB)
  }
}
save(uniform.se.res, file = "uniform.se.res.RData")

#Table SE.F
#beta.se.res <- list()
#nrep <- 500
#nvar <- 1
#coef <- c(1,-1)
#for(n in c(50,100,500,1000)){
#  name <- paste("n= ",n, sep="")
#  print(name)
#  beta.se.res[[name]] <- sim_std_error(nrep = nrep, n = n, nvar = nvar, coef = coef, distribution = "beta", bandwidth = "normal", h.SJ = TRUE)
#  print(beta.se.res[[name]]$summary)
#  print(beta.se.res[[name]]$SE1$ratio1)
#  print(beta.se.res[[name]]$SE1A$ratio1A)
#  print(beta.se.res[[name]]$SE2$ratio2)
#  print(beta.se.res[[name]]$SE2A$ratio2A)
#  if (n <= 100){
#    print(beta.se.res[[name]]$SEB$ratioB)
#  }
#}
#save(beta.se.res, file = "beta.se.res.RData")

#Table SE.F
gamma.se.res <- list()
nrep <- 500
nvar <- 1
coef <- c(1,-1)
for(n in c(50,100,500,1000)){
  name <- paste("n= ",n, sep="")
  print(name)
  gamma.se.res[[name]] <- sim_std_error(nrep = nrep, n = n, nvar = nvar, coef = coef, distribution = "gamma", bandwidth = "robust2", h.SJ = TRUE)
  print(gamma.se.res[[name]]$summary)
  print(gamma.se.res[[name]]$SE1$ratio1)
  print(gamma.se.res[[name]]$SE1A$ratio1A)
  print(gamma.se.res[[name]]$SE2$ratio2)
  print(gamma.se.res[[name]]$SE2A$ratio2A)
  if (n <= 100){
    print(gamma.se.res[[name]]$SEB$ratioB)
  }
}
save(gamma.se.res, file = "gamma.se.res.RData")

#Table SE.G
skewed.se.res <- list()
nrep <- 500
nvar <- 1
coef <- c(1,-1)
for(n in c(50,100,500,1000)){
  name <- paste("n= ",n, sep="")
  print(name)
  skewed.se.res[[name]] <- sim_std_error(nrep = nrep, n = n, nvar = nvar, coef = coef, distribution = "skewed", bandwidth = "robust2", h.SJ = TRUE)
  print(skewed.se.res[[name]]$summary)
  print(skewed.se.res[[name]]$SE1$ratio1)
  print(skewed.se.res[[name]]$SE1A$ratio1A)
  print(skewed.se.res[[name]]$SE2$ratio2)
  print(skewed.se.res[[name]]$SE2A$ratio2A)
  if (n <= 100){
    print(skewed.se.res[[name]]$SEB$ratioB)
  }
}
save(skewed.se.res, file = "skewed.se.res.RData")

#Table SE.H
log_normal.se.res <- list()
nrep <- 500
nvar <- 1
coef <- c(1,-1)
for(n in c(50,100,500,1000)){
  name <- paste("n= ",n, sep="")
  print(name)
  log_normal.se.res[[name]] <- sim_std_error(nrep = nrep, n = n, nvar = nvar, coef = coef, distribution = "log_normal", bandwidth = "robust2", h.SJ = TRUE)
  print(log_normal.se.res[[name]]$summary)
  print(log_normal.se.res[[name]]$SE1$ratio1)
  print(log_normal.se.res[[name]]$SE1A$ratio1A)
  print(log_normal.se.res[[name]]$SE2$ratio2)
  print(log_normal.se.res[[name]]$SE2A$ratio2A)
  if (n <= 100){
    print(log_normal.se.res[[name]]$SEB$ratioB)
  }
}
save(log_normal.se.res, file = "log_normal.se.res.RData")

